@@ -1,3 +1,4 @@
+import inspect
 import io
 import os
 import random
@@ -16,7 +17,7 @@ from speckenv_django import django_database_url
 from fh_fablib.extract_js_gettext_strings import generate_strings
 
 
-__version__ = "1.0.20220824"
+__version__ = "1.0.20230303"
 
 
 # I don't care, in this context.
@@ -30,7 +31,8 @@ def ansi(code):
 # underline = ansi("4")
 red = ansi("31")
 green = ansi("32")
-purple = ansi("35")
+yellow = ansi("33")
+magenta = ansi("35")
 
 
 def progress(msg):
@@ -38,7 +40,11 @@ def progress(msg):
 
 
 def info(msg):
-    print(purple(msg))
+    print(magenta(msg))
+
+
+def warning(msg):
+    print(yellow(msg))
 
 
 def terminate(msg):
@@ -46,6 +52,31 @@ def terminate(msg):
     sys.exit(1)
 
 
+def _bool(s):
+    true = {"yes", "true"}
+    false = {"no", "false"}
+    if s in true:
+        return True
+    elif s in false:
+        return False
+    terminate(
+        f"Boolean argument value {s!r} not recognized. Use yes, true, no or false."
+    )
+
+
+def _find_base():
+    frame = inspect.currentframe().f_back.f_back
+    try:
+        while frame:
+            if (name := frame.f_locals.get("__file__")) and name.endswith(
+                "/fabfile.py"
+            ):
+                return Path(name).parent
+            frame = frame.f_back
+    finally:
+        del frame
+
+
 def require(version):
     if __version__ < version:
         terminate(f"fh_fablib version {version} required (you have {__version__})")
@@ -60,25 +91,24 @@ def require(version):
             )
             if new != old:
                 path.write_text(new)
-                info(
-                    "The fabfile has been updated automatically with your"
-                    " local fh_fablib version."
+                _update_dotfiles(force=True)
+                warning(
+                    "The fabfile and dotfiles have been updated automatically.\n\nPlease check the result twice before committing!"
                 )
-                return
-
-        info(f"fh_fablib version is {__version__}, project requires only {version}.")
 
 
 def run(c, *a, **kw):
     """A Context.run or Connection.run with better defaults"""
-    kw.setdefault("echo", True)
     kw.setdefault("pty", True)
     kw.setdefault("replace_env", False)
+    if not kw.get("hide"):
+        progress(" ".join(str(part) for part in a))
     return c.run(*a, **kw)
 
 
 class Config:
     app = "app"
+    base = _find_base()
     environment = "default"
     environments = {}
     force = False
@@ -87,8 +117,6 @@ class Config:
         for key, value in kwargs.items():
             setattr(self, key, value)
 
-        os.chdir(self.base)
-
     def __getattr__(self, key):
         environments = getattr(self, "environments", None)
         if environments:
@@ -102,11 +130,15 @@ class Config:
 
 #: Defaults
 config = Config()
+os.chdir(config.base)
 
 
 def environment(name, cfg, **kwargs):
     config.environments[name] = cfg
 
+    if name in kwargs.get("aliases", ()):
+        terminate(f"Remove {name} from the aliases list of the {name} environment.")
+
     @task(name=name, **kwargs)
     def fn(ctx):
         cfg["environment"] = name
@@ -164,18 +196,21 @@ for job in $(jobs -p); do wait $job; done
         run(ctx, f"bash {f.name}", replace_env=False)
 
 
-@task(auto_shortflags=False, help={"force": "Overwrite existing pre-commit files"})
-def hook(ctx, force=False):
-    """
-    Add default pre-commit configuration and install hook running coding style checks
-    """
+def _update_dotfiles(*, force):
     source = Path(__file__).parent / "dotfiles"
     target = config.base
     for s in sorted(source.glob("*")):
         t = target / s.name
         if force or not t.exists():
             shutil.copy(s, t)
-            info(f"Copying {s.name} into project...")
+
+
+@task(auto_shortflags=False, help={"force": "Overwrite existing pre-commit files"})
+def hook(ctx, force=False):
+    """
+    Add default pre-commit configuration and install hook running coding style checks
+    """
+    _update_dotfiles(force=force)
     run(ctx, "pre-commit install -f")
 
 
@@ -333,7 +368,7 @@ def cm(ctx):
 
 
 def _python3():
-    interpreters = ("python3.10", "python3.9", "python3.8")
+    interpreters = ("python3.11", "python3.10", "python3.9", "python3.8")
     return next(filter(None, (shutil.which(v) for v in interpreters)))
 
 
@@ -341,11 +376,18 @@ def _pip_up(c):
     run(c, "venv/bin/python -m pip install -U pip wheel setuptools")
 
 
-@task(auto_shortflags=False, help={"stable": "Avoid pre-release versions of packages"})
-def upgrade(ctx, stable=False):
+@task(
+    auto_shortflags=False,
+    help={
+        "keep": "Keep the existing virtualenv",
+        "stable": "Avoid pre-release versions of packages",
+    },
+)
+def upgrade(ctx, keep=False, stable=False):
     """Re-create the virtualenv with newest versions of all libraries"""
-    run(ctx, "rm -rf venv")
-    run(ctx, f"{_python3()} -m venv venv")
+    venv = config.base / "venv"
+    if not venv.exists() or not keep:
+        run(ctx, f"rm -rf venv && {_python3()} -m venv venv")
     _pip_up(ctx)
     extra = "" if stable else "--pre"
     run(ctx, f"venv/bin/python -m pip install -U -r requirements-to-freeze.txt {extra}")
@@ -371,10 +413,15 @@ def update(ctx):
     if not venv.exists():
         run(ctx, f"{_python3()} -m venv venv")
     _pip_up(ctx)
-    run(ctx, "venv/bin/python -m pip install -r requirements.txt")
-    run(ctx, "git submodule update --init")
-    run(ctx, 'find . -name "*.pyc" -delete')
-    run(ctx, "yarn")
+    _concurrently(
+        ctx,
+        [
+            "venv/bin/python -m pip install -r requirements.txt",
+            "git submodule update --init",
+            'find . -name "*.pyc" -delete',
+            "yarn",
+        ],
+    )
     run(ctx, "venv/bin/python manage.py migrate", warn=True)
     hook(ctx)
 
@@ -434,8 +481,9 @@ def nine_vhost(ctx):
 
 
 @task(auto_shortflags=False, help={"include-www": "Include the www. subdomain"})
-def nine_alias_add(ctx, alias, include_www=False):
+def nine_alias_add(ctx, alias, include_www):
     """Add aliasses to a nine-manage-vhost virtual host"""
+    include_www = _bool(include_www)
     with Connection(config.host) as conn:
         run(
             conn,
@@ -452,8 +500,9 @@ def nine_alias_add(ctx, alias, include_www=False):
 
 
 @task(auto_shortflags=False, help={"include-www": "Include the www. subdomain"})
-def nine_alias_remove(ctx, alias, include_www=False):
+def nine_alias_remove(ctx, alias, include_www):
     """Remove aliasses from a nine-manage-vhost virtual host"""
+    include_www = _bool(include_www)
     with Connection(config.host) as conn:
         run(
             conn,
@@ -469,31 +518,66 @@ def nine_alias_remove(ctx, alias, include_www=False):
             )
 
 
+def _unit(config, *, args=""):
+    # args = " -w 2 --preload"
+    return f"""\
+[Unit]
+Description=gunicorn for {config.domain}
+
+[Service]
+Environment=LANG=en_US.UTF-8 LC_ALL=en_US.UTF-8 LC_CTYPE=en_US.UTF-8
+ExecStart=/home/www-data/{config.domain}/venv/bin/gunicorn wsgi:application -b unix:///home/www-data/{config.domain}/tmp/gunicorn.sock --max-requests 1000 --max-requests-jitter 100 {args}
+SyslogIdentifier=gunicorn:{config.domain}
+WorkingDirectory=/home/www-data/{config.domain}/
+Restart=always
+
+[Install]
+WantedBy=default.target
+"""
+
+
 @task
 def nine_unit(ctx):
-    """Start and enable a gunicorn@ unit"""
+    """Start and enable a gunicorn unit"""
     with Connection(config.host) as conn:
-        run(conn, f"systemctl --user enable --now gunicorn@{config.domain}.service")
+        conn.put(
+            io.StringIO(_unit(config)), f".config/systemd/user/{config.domain}.service"
+        )
+        run(conn, "systemctl --user daemon-reload")
+        run(conn, f"systemctl --user enable --now {config.domain}.service")
+
+    info("Successfully created the virtual host.\n")
+    warning("Please update the hostings overview as well!")
 
 
 def _nine_has_manage_databases(conn):
     return bool(run(conn, "which nine-manage-databases", warn=True).stdout.strip())
 
 
-@task
-def nine_db_dotenv(ctx):
+@task(
+    auto_shortflags=False,
+    help={"recreate": "Only recreate the database"},
+)
+def nine_db_dotenv(ctx, recreate=False):
     """Create a database and initialize the .env"""
     with Connection(config.host) as conn:
-        try:
-            conn.get(f"{config.domain}/.env", io.BytesIO())
-        except FileNotFoundError:
-            pass
-        else:
-            terminate(f"'{config.domain}/.env' already exists on the server")
+        if recreate:
+            e = _srv_env(conn, f"{config.domain}/.env")
+            db = django_database_url(e("DATABASE_URL", required=True))
 
-        password = _random_string(20, chars="abcdefghijklmnopqrstuvwxyz0123456789")
-        secret_key = _random_string(70)
-        dbname = _dbname_from_domain(config.domain)
+            password = db["PASSWORD"]
+            dbname = db["NAME"]
+        else:
+            try:
+                conn.get(f"{config.domain}/.env", io.BytesIO())
+            except FileNotFoundError:
+                pass
+            else:
+                terminate(f"'{config.domain}/.env' already exists on the server")
+
+            password = _random_string(20, chars="abcdefghijklmnopqrstuvwxyz0123456789")
+            secret_key = _random_string(70)
+            dbname = _dbname_from_domain(config.domain)
 
         if _nine_has_manage_databases(conn):
             dbname = f"nmd_{dbname}"
@@ -521,9 +605,10 @@ def nine_db_dotenv(ctx):
                 f'"',
             )
 
-        conn.put(
-            io.StringIO(
-                f"""\
+        if not recreate:
+            conn.put(
+                io.StringIO(
+                    f"""\
 DEBUG=False
 DATABASE_URL=postgres://{dbname}:{password}@localhost:5432/{dbname}
 CACHE_URL=hiredis://localhost:6379/1/?key_prefix={dbname}
@@ -540,9 +625,9 @@ SENTRY_ENVIRONMENT=
 # SECURE_SSL_HOST={config.domain}
 # SECURE_SSL_REDIRECT=True
 """
-            ),
-            f"{config.domain}/.env",
-        )
+                ),
+                f"{config.domain}/.env",
+            )
 
 
 @task
@@ -562,7 +647,7 @@ def nine_ssl(ctx):
 
 
 def _nine_restart(conn):
-    run(conn, f"systemctl --user restart gunicorn@{config.domain}.service")
+    run(conn, f"systemctl --user restart {config.domain}.service")
 
 
 @task
@@ -575,10 +660,13 @@ def nine_restart(ctx):
 
 @task
 def nine_disable(ctx):
-    """Disable a virtual host, dump and remove the DB and stop the gunicorn@ unit"""
+    """Disable a virtual host, dump and remove the DB and stop the gunicorn unit"""
     with Connection(config.host) as conn:
         run(conn, f"sudo nine-manage-vhosts virtual-host remove {config.domain}")
-        run(conn, f"systemctl --user disable --now gunicorn@{config.domain}.service")
+
+        run(conn, f"systemctl --user disable --now {config.domain}.service")
+        run(conn, f"rm -f .config/systemd/user/{config.domain}.service")
+        run(conn, "systemctl --user daemon-reload")
 
         e = _srv_env(conn, f"{config.domain}/.env")
         srv_dsn = e("DATABASE_URL")
@@ -592,6 +680,9 @@ def nine_disable(ctx):
             run(conn, f"source ~/.profile && dropdb {srv_dbname}")
             run(conn, f"source ~/.profile && dropuser {srv_dbname}")
 
+    info("Successfully removed the virtual host.\n")
+    warning("Please update the hostings overview as well!")
+
 
 @task
 def nine_checkout(ctx):
@@ -672,38 +763,6 @@ def nine(ctx):
     nine_db_dotenv(ctx)
     nine_vhost(ctx)
     nine_unit(ctx)
-    # nine_ssl(ctx)      Does not apply
-    # nine_disable(ctx)  Does obviously not apply 😅
-
-
-@task
-def bitbucket(ctx):
-    """Create a repository on Bitbucket and push the code"""
-    e = _local_env(Path.home() / ".box.env")
-    username = e("BITBUCKET_USERNAME")
-    password = e("BITBUCKET_PASSWORD")
-    organization = e("BITBUCKET_ORGANIZATION")
-    repository = config.domain
-
-    print(f"Username [{username}]: ", end="")
-    username = input() or username
-    print(f"Password [{password}]: ", end="")
-    password = input() or password
-    print(f"Organization [{organization}]: ", end="")
-    organization = input() or organization
-    print(f"Repository [{repository}]: ", end="")
-    repository = input() or repository
-
-    run(
-        ctx,
-        f"""\
-curl -X POST -v -u {username}:"{password}" -H "content-type: application/json"\
- https://api.bitbucket.org/2.0/repositories/{organization}/{repository}\
- -d '{{"scm": "git", "is_private": true, "forking_policy": "no_public_forks"}}'\
-""",
-    )
-    run(ctx, f"git remote add origin git@bitbucket.org:{organization}/{repository}.git")
-    run(ctx, f"git push -u origin {config.branch}")
 
 
 @task
@@ -722,7 +781,10 @@ def github(ctx):
     print(f"Repository [{repository}]: ", end="")
     repository = input() or repository
 
-    run(ctx, f"gh repo create {organization}/{repository} --private")
+    run(
+        ctx,
+        f"gh repo create {organization}/{repository} --private --source=. --remote=origin --push",
+    )
     run(ctx, f"git push -u origin {config.branch}")
 
 
@@ -828,7 +890,6 @@ GENERAL = {
     reset_pw,
     reset_sq,
     local,
-    bitbucket,
     github,
     fetch,
     check,
